<?php
/**
 * @author 河南鼎汉软件科技
 * @copyright Copyright (c) 2017 HNDH Software Technology Co., Ltd.
 * createtime: 2017/10/12 22:36
 */

namespace common\components;

use common\models\Config;
use common\models\TradeMatch;
use common\models\TradeOrder;
use yii\base\BaseObject;
use yii\base\Exception;

/**
 * Class Matchmaking
 * @package common\components
 */
class Matchmaking extends BaseObject
{
    const BUY_KEY_PREFIX = 'deal_buy_';
    const SELL_KEY_PREFIX = 'deal_sell_';
    const LOCK_PREFIX = 'deal_lock_';
    const DELETED = '_DELETED_';

    /**
     * @var \yii\redis\Connection redis实例化对象
     */
    public $redis;

    /**
     * @var array 交易订单
     */
    public $order;

    /**
     * @var integer 交易类型 1买9卖
     */
    public $dealType;

    /**
     * @var array 匹配结果集
     */
    public $match = [];

    /**
     * @var array 交易规则
     */
    public $rules;

    public function init()
    {
        $this->redis = \Yii::$app->get('redis');
        $this->rules = json_decode(Config::getConfig('TRADE_RULES'));
        $this->dealType = $this->order['type'];
    }

    /**
     * 交易匹配
     * 1. 匹配成功的出队，未完成交易数量的进入等待队列（买入|卖出）
     * @return bool
     * @throws Exception
     */
    public function run()
    {
        \Yii::warning('开始匹配' . $this->order['id'], 'DEAL');
        if ($this->dealType == '1') {
            $queueKeys = $this->redis->keys(self::SELL_KEY_PREFIX . '*');
            sort($queueKeys);
            foreach ($queueKeys as $key) {
                if ($this->order['num'] <= $this->order['deal_num'] || $this->getPriceFromKey($key, self::SELL_KEY_PREFIX) > $this->order['sell_price']) {
                    break;
                }
                $matchOrder = $this->matchOrder($key);

                // 释放锁
                $this->unlock(self::LOCK_PREFIX . $key);
                if ($matchOrder === true) {
                    break;
                }
            }
            // 未完成交易进入买入队列

            \Yii::warning('判断是否进入队列' . json_encode($this->order), 'DEAL');
            if ($this->order['status'] == TradeOrder::DEAL_STATUS) {
                $result = $this->redis->rpush(self::BUY_KEY_PREFIX . sprintf('%.2f', $this->order['sell_price']), json_encode($this->order));
                if (empty($result)) {
                    \Yii::error('匹配异常，请重试' . json_encode($this->order), 'DEAL');
                    throw new Exception('匹配异常，请重试');
                }
                \Yii::warning('进入队列' . json_encode($this->order), 'DEAL');
            }
            return true;
        } else if ($this->dealType == '9') {
            $queueKeys = $this->redis->keys(self::BUY_KEY_PREFIX . '*');
            rsort($queueKeys);
            foreach ($queueKeys as $key) {
                if ($this->order['num'] <= $this->order['deal_num'] || $this->getPriceFromKey($key, self::BUY_KEY_PREFIX) < $this->order['sell_price']) {
                    break;
                }
                $matchOrder = $this->matchOrder($key);

                // 释放锁
                $this->unlock(self::LOCK_PREFIX . $key);
                if ($matchOrder === true) {
                    break;
                }
            }
            // 未完成交易进入卖出队列
            \Yii::warning('判断是否进入队列' . json_encode($this->order), 'DEAL');
            if ($this->order['status'] == TradeOrder::DEAL_STATUS) {
                $result = $this->redis->rpush(self::SELL_KEY_PREFIX . sprintf('%.2f', $this->order['sell_price']), json_encode($this->order));
                if (empty($result)) {
                    \Yii::error('匹配异常，请重试' . json_encode($this->order), 'DEAL');
                    throw new Exception('匹配异常，请重试');
                }
                \Yii::warning('进入队列' . json_encode($this->order), 'DEAL');
            }
            return true;
        } else {
            return false;
        }
    }

    /**
     * 处理买入匹配订单
     * 如果匹配的订单还满足数量需求就返回false要求继续匹配
     * @param string $key 匹配到的交易价订单列表键值
     * @return bool 不需要继续匹配返回true，否则返回false
     */
    private function matchOrder($key)
    {
        $length = $this->redis->llen($key);
        if (empty($length)) {
            return false;
        }

        // 等待锁释放
        while (!$this->lock(self::LOCK_PREFIX . $key)) {
            usleep(100);
        }

        for ($i = 0; $i < $length; $i++) {
            $matchOrder = $this->redis->lindex($key, $i);
            if (empty($matchOrder) || $matchOrder == self::DELETED) {
                continue;
            }
            $this->redis->lset($key, $i, self::DELETED); // 防止并发读取同一个列表数据

            \Yii::warning('匹配详情' . json_encode($this->order) . $matchOrder, 'DEAL');
            $matchOrder = json_decode($matchOrder, true);
            if (($matchOrder['num'] - $matchOrder['deal_num']) > ($this->order['num'] - $this->order['deal_num'])) {
                $dealNum = $this->order['num'] - $this->order['deal_num'];
                $matchOrder['deal_num'] += $dealNum;
                $this->order['deal_num'] += $dealNum;
                $this->order['status'] = TradeOrder::DEAL_OK_STATUS;
                $this->addOkQueue($matchOrder, $dealNum);
                $this->redis->lset($key, $i, json_encode($matchOrder));
                return true;
            } else if (($matchOrder['num'] - $matchOrder['deal_num']) < ($this->order['num'] - $this->order['deal_num'])) {
                $dealNum = $matchOrder['num'] - $matchOrder['deal_num'];
                $this->order['deal_num'] += $dealNum;
                $matchOrder['deal_num'] += $dealNum;
                $matchOrder['status'] = TradeOrder::DEAL_OK_STATUS;
                $this->addOkQueue($matchOrder, $dealNum);
                $this->redis->lrem($key, 0, self::DELETED);
            } else {
                $dealNum = $this->order['num'] - $this->order['deal_num'];
                $matchOrder['deal_num'] += $dealNum;
                $this->order['deal_num'] += $dealNum;
                $this->order['status'] = 9;
                $matchOrder['status'] = 9;
                $this->addOkQueue($matchOrder, $dealNum);
                $this->redis->lrem($key, 0, self::DELETED);
                return true;
            }
        }

        return false;
    }

    /**
     * 获取key值中的价格
     * @param $str
     * @param $prefix
     * @return float
     * @throws Exception
     */
    private function getPriceFromKey($str, $prefix)
    {
        $pos = strrpos($str, $prefix);
        if ($pos === false) {
            throw new Exception('没有找到对应的前缀！');
        }
        $price = substr($str, $pos + strlen($prefix));
        if (empty($price)) {
            throw new Exception('没有找到价格信息！');
        }

        return sprintf("%.2f", $price);
    }

    /**
     * 订单匹配结果OK出队列
     * @param $matchOrder
     * @param $deal_num
     * @throws Exception
     */
    private function addOkQueue($matchOrder, $deal_num)
    {
        $rules = json_decode(Config::getConfig('TRADE_RULES'), true);
        $buyFee = bcmul($deal_num * $matchOrder['sell_price'], $rules['buy']['fee'] * 0.01, 2);
        $sellFee = bcmul($deal_num * $matchOrder['sell_price'], $rules['sell']['fee'] * 0.01, 2);

        $match = new TradeMatch();
        $match->type = $this->dealType;
        $match->uid = $this->order['uid'];
        $match->order_id = $this->order['id'];
        if ($this->dealType == '1') {
            $match->fee = $buyFee;
        } else {
            $match->fee = $sellFee;
        }
        $match->status = $this->order['status'];
        $match->m_uid = $matchOrder['uid'];
        $match->m_order_id = $matchOrder['id'];
        if ($this->dealType == '1') {
            $match->m_fee = $sellFee;
        } else {
            $match->m_fee = $buyFee;
        }
        $match->m_status = $matchOrder['status'];
        $match->deal_price = $matchOrder['sell_price'];
        $match->deal_num = $deal_num;
        $match->is_handle = 0;
        if (!$match->save()) {
            throw new Exception('订单匹配记录添加失败！');
        }
        $this->match[] = $match->id;
    }

    /**
     * 获取锁
     * @param  String $key 锁标识
     * @param  Int $expire 锁过期时间
     * @return bool
     */
    public function lock($key, $expire = 3)
    {
        $is_lock = $this->redis->setnx($key, time() + $expire);

        // 不能获取锁
        if (!$is_lock) {

            // 判断锁是否过期
            $lock_time = $this->redis->get($key);

            // 锁已过期，删除锁，重新获取
            if (time() > $lock_time) {
                $this->unlock($key);
                $is_lock = $this->redis->setnx($key, time() + $expire);
            }
        }

        return $is_lock ? true : false;
    }

    /**
     * 释放锁
     * @param  String $key 锁标识
     * @return bool
     */
    public function unlock($key)
    {
        return $this->redis->del($key);
    }
}
